/*--------------------------------------------------------------------
  File:        xgraphics.c
  Author:      Nils Schmeisser
  ----------------------------------------------------------------------
  Description: Graphics module for X applications
  ----------------------------------------------------------------------
  Notes:       
  Date:        3.12.1993
  ----------------------------------------------------------------------
   Copyright (C) 1993 Nils Schmeisser

This file is part of libxapp.

libxapp is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

libxapp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with libxapp; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/ 

#define XTFUNCPROTO
#define ANSI

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// on Convex, HP-UX replace this by the next line
//#include <values.h>
#include <sys/param.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/X.h>
#include <Xm/DrawingA.h>
extern Widget xCreatePopupShell(char*,int,int,int,int,Widget);
extern Widget xCreateDrawingArea(int width,int height,Widget parent);
extern void xRegisterWindow(Window w);

typedef void (*XGraphicsRedrawFunction)(Window,int);
typedef XPoint polypoint;

int xDARKBLUE;
int xDARKGREEN;
int xDARKCYAN;
int xDARKRED;
int xDARKMAGENTA;
int xDARKBROWN;
int xBLACK;
int xBLUE;
int xGREEN;
int xCYAN;
int xRED;
int xMAGENTA;
int xBROWN;
int xLIGHTGRAY;
int xDARKGRAY;
int xLIGHTBLUE;
int xLIGHTGREEN;
int xLIGHTCYAN;
int xLIGHTRED;
int xLIGHTMAGENTA;
int xYELLOW;
int xWHITE;

typedef struct GW {
          Window window;
          float ratio;
          int  x,y,width,height;
          int  foreground,background;
          int  penx,peny;
          int  fill_pattern;
          XGraphicsRedrawFunction redraw;
          GW   *next;
        };

static int ColorOfs=40;
static Colormap ColorTable;
static XColor   ColorEntry[256];
static Window root;
static Window default_window;
static int screen;
static Visual *VisualInfo;
static int LMAXCOLORS=256;
static Display *GRAPHICS_DISPLAY=NULL;
static GC GRAPHICS_CONTEXT;

static GW *GW_ROOT=NULL;
static GW *default_gw=NULL;

//=== Private, handling for Windows information ===================================

static GW *FIND_GW(Window window) {
  GW *p;

  if (default_gw!=NULL) if (default_gw->window!=window) default_gw=NULL;
  if (default_gw==NULL) {
    p=GW_ROOT;
    while ((p!=NULL) && (p->window!=window)) p=p->next;
    default_gw=p;
    return(p);
  } else return(default_gw);
}

static void NEW_GW(Window window,int x,int y,int w,int h) {
  GW *p;

  if (GW_ROOT==NULL) {
    GW_ROOT=new GW;
    p=GW_ROOT;
  } else {
    p=GW_ROOT;
    while (p->next!=NULL) p=p->next;
    p->next=new GW;
    p=p->next;
  }
  p->window=window;
  p->x=x;
  p->y=y;
  p->width=w;
  p->height=h;
  p->next=NULL;
}

static void DEL_GW(Window window) {
  GW *p,*pw;

  pw=FIND_GW(window);
  if (pw==GW_ROOT) {
    if (pw->next!=NULL) GW_ROOT=pw->next; else GW_ROOT=NULL;
    delete pw;
  } else {
    p=GW_ROOT;
    while (p->next!=pw) p=p->next;
    p->next=pw->next;
    delete pw;
  }  
}

static void SET_foreground(Window window,int color) {
  FIND_GW(window)->foreground=color;
}

static void SET_background(Window window,int color) {
  FIND_GW(window)->background=color;
}

static void SET_pen(Window window,int x,int y) {
  GW *p;

  p=FIND_GW(window);
  p->penx=x;
  p->peny=y;
}

static int GET_foreground(Window window) {
  return(FIND_GW(window)->foreground);
}

static int GET_background(Window window) {
  return(FIND_GW(window)->background);
}

static void GET_pen(Window window,int &x,int &y) {
  GW *p;

  p=FIND_GW(window);
  x=p->penx;
  y=p->peny;
}

static int ConvRGBtoPIX(int red,int green,int blue) {
  int i,c;
  int dist,min=256*256*3+1;

  for (i=0;i<256;i++) {
    dist= abs((int)ColorEntry[i].red/256-red)
         +abs((int)ColorEntry[i].green/256-green)
         +abs((int)ColorEntry[i].blue/256-blue);
    if (dist<min) {
      min=dist;
      c=i;
    }
  }
  return(c);
}

static void create_standard_colors() {
  xBLACK       =ConvRGBtoPIX(  0,  0,  0)-ColorOfs;
  xDARKGRAY    =ConvRGBtoPIX( 63, 63, 63)-ColorOfs;
  xDARKBLUE    =ConvRGBtoPIX(  0,  0,127)-ColorOfs;
  xDARKGREEN   =ConvRGBtoPIX(  0,127,  0)-ColorOfs;
  xDARKCYAN    =ConvRGBtoPIX(  0,127,127)-ColorOfs;
  xDARKRED     =ConvRGBtoPIX(127,  0,  0)-ColorOfs;
  xDARKMAGENTA =ConvRGBtoPIX(127,  0,127)-ColorOfs;
  xDARKBROWN   =ConvRGBtoPIX(127,127,  0)-ColorOfs;

  xBLUE        =ConvRGBtoPIX(  0,  0,255)-ColorOfs;
  xGREEN       =ConvRGBtoPIX(  0,255,  0)-ColorOfs;
  xCYAN        =ConvRGBtoPIX(  0,255,255)-ColorOfs;
  xRED         =ConvRGBtoPIX(255,  0,  0)-ColorOfs;
  xMAGENTA     =ConvRGBtoPIX(255,  0,255)-ColorOfs;
  xBROWN       =ConvRGBtoPIX(255,255,  0)-ColorOfs;

  xLIGHTGRAY   =ConvRGBtoPIX(127,127,127)-ColorOfs;
  xLIGHTBLUE   =ConvRGBtoPIX(  0,  0,255)-ColorOfs;
  xLIGHTGREEN  =ConvRGBtoPIX(  0,255,  0)-ColorOfs;
  xLIGHTCYAN   =ConvRGBtoPIX(  0,255,255)-ColorOfs;
  xLIGHTRED    =ConvRGBtoPIX(255,  0,  0)-ColorOfs;
  xLIGHTMAGENTA=ConvRGBtoPIX(255,  0,255)-ColorOfs;
  xYELLOW      =ConvRGBtoPIX(255,255,  0)-ColorOfs;
  xWHITE       =ConvRGBtoPIX(255,255,255)-ColorOfs;
}

//=================================================================================

Display *getgraphicsdisplay() { return(GRAPHICS_DISPLAY); }

void setgraphicsdisplay(Display *dpy) { GRAPHICS_DISPLAY=dpy; }

void setwindow(Window window) { 
  XWindowAttributes Attr;
  GW *p;

  default_window=window;
  p=FIND_GW(default_window);

  if (p!=NULL) {
    XGetWindowAttributes(GRAPHICS_DISPLAY,window,&Attr);
    p->height=Attr.height;
    p->width=Attr.width;  
    p->x=Attr.x;
    p->y=Attr.y;
    XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,p->foreground+ColorOfs);
    XSetWindowBackground(GRAPHICS_DISPLAY,window,p->background+ColorOfs);
  } else {
    printf("Window %d is not in graphics window list.\n",window);
  }
}

Window openwindow(char *name,int x,int y,int sizex,int sizey,Widget parent,Widget *da=NULL) {
  XSetWindowAttributes Attr;
  Window window;
  Widget gshell,drawarea;
  Arg args[10];
  int n=0;

  gshell=xCreatePopupShell(name,x,y,sizex,sizey,parent);
    drawarea=xCreateDrawingArea(sizex,sizey,gshell);
  XtPopup(gshell,XtGrabNonexclusive);
  window=XtWindow(gshell);
  Attr.event_mask=ExposureMask | ColormapChangeMask | StructureNotifyMask | 
                  ButtonPressMask | KeyPressMask;
  Attr.colormap  =ColorTable;
  XChangeWindowAttributes(GRAPHICS_DISPLAY,window,CWEventMask | CWColormap,&Attr);  
  XMapWindow(GRAPHICS_DISPLAY,window);
  XSelectInput(GRAPHICS_DISPLAY,window,
               ColormapChangeMask | ExposureMask | StructureNotifyMask |
               ButtonPressMask | KeyPressMask);
  XFlush(GRAPHICS_DISPLAY);
  NEW_GW(window,x,y,sizex,sizey);
  default_window=window;
  xRegisterWindow(window);
  if (da!=NULL) *da=drawarea;
  return(window);
}

void closewindow(Window window) {
  DEL_GW(window);

  if (window==default_window) {
    if (GW_ROOT!=NULL) default_window=GW_ROOT->window;
      else default_window=MAXINT;
  }
  XDestroyWindow(GRAPHICS_DISPLAY,window);
}

void cleardevice(Window);

void RedrawXGraphicsWindow(Widget win,XtPointer,XEvent *event,Boolean*) {
  XWindowAttributes Attr;
  Window d;
  GW *p;
  int reason;
  XEvent cevent;
  Boolean resize=False;
  XExposeEvent        expose;

  d=XtWindow(win);
  XFlush(GRAPHICS_DISPLAY);
  switch (event->type) {
        case ButtonPress: reason=ButtonPress;
                          break;
           case KeyPress: reason=KeyPress;
                          break;
             case Expose: expose   =event->xexpose;
                          reason=Expose;
                          while (expose.count!=0) {
                            XNextEvent(GRAPHICS_DISPLAY,event);
                            expose=event->xexpose;
                          }
                          
                          // looking ahead if next event in the queue is a resize
                          
                          if (XPending(GRAPHICS_DISPLAY)>0) {
                            XPeekEvent(GRAPHICS_DISPLAY,&cevent);                                
                            if (cevent.type==ResizeRequest) {
                              reason=ResizeRequest;
                              XNextEvent(GRAPHICS_DISPLAY,&cevent);
                            }
                          }
                          break;
    case ConfigureNotify: reason=ResizeRequest;
                          
                          // if next event in the queue is an expose then remove event
                          
                          if (XPending(GRAPHICS_DISPLAY)>0) {
                            XPeekEvent(GRAPHICS_DISPLAY,&cevent);
                            if (cevent.type==Expose) XNextEvent(GRAPHICS_DISPLAY,&cevent);    
                            while (cevent.xexpose.count!=0) XNextEvent(GRAPHICS_DISPLAY,&cevent);
                          }
  }

  p=FIND_GW(d);
  XGetWindowAttributes(GRAPHICS_DISPLAY,d,&Attr);
  p->height=Attr.height;
  p->width=Attr.width;  
  p->x=Attr.x;
  p->y=Attr.y;

  if (reason==ResizeRequest) {
    if (p->ratio>0) {
      if (Attr.height!=(int)((float)Attr.height*p->ratio)) {
        p->width=(int)((float)Attr.height*p->ratio);
        XResizeWindow(GRAPHICS_DISPLAY,d,p->width,p->height);
      }
    }
    cleardevice(d);
  }
  p->redraw(d,reason);
}

void setredrawfunction(Window window,XGraphicsRedrawFunction redraw_function) {
  Widget draw;

  draw=XtWindowToWidget(GRAPHICS_DISPLAY,window);
  XtAddEventHandler(draw,
                    ExposureMask | StructureNotifyMask | ButtonPressMask | KeyPressMask,
                    False,RedrawXGraphicsWindow,NULL);
  FIND_GW(window)->redraw=redraw_function;
}

Window openwindow(char *name,int x,int y,int sizex,int sizey,
                  Widget parent,XGraphicsRedrawFunction redraw_function,Widget *da=NULL) {

  Window window=openwindow(name,x,y,sizex,sizey,parent,da);
  setredrawfunction(window,redraw_function);
  return window;
}

//---------------------------------------------------------------------------------

void setrgbpalette(int colornum, int red, int green, int blue) {
  ColorEntry[colornum+ColorOfs].pixel  =(unsigned long)(colornum+ColorOfs);
  ColorEntry[colornum+ColorOfs].red    =(unsigned short)(red*256);
  ColorEntry[colornum+ColorOfs].green  =(unsigned short)(green*256);
  ColorEntry[colornum+ColorOfs].blue   =(unsigned short)(blue*256);
  ColorEntry[colornum+ColorOfs].flags  = DoRed | DoGreen | DoBlue;    
  XStoreColor(GRAPHICS_DISPLAY,ColorTable,&ColorEntry[colornum+ColorOfs]);
  XFlush(GRAPHICS_DISPLAY);
}

void cleardevice(Window window=-MAXINT) {
  if (window==-MAXINT) XClearWindow(GRAPHICS_DISPLAY,default_window); 
    else XClearWindow(GRAPHICS_DISPLAY,window);
}

//---------------------------------------------------------------------------------

void initgraph(Display *display=NULL) {
  char driver[40];
  int i;
  XColor q;
  XGCValues GCValues;

// get number of x,y,colors

  if (display==NULL) GRAPHICS_DISPLAY=XOpenDisplay((char*)NULL); else GRAPHICS_DISPLAY=display;
  screen=XDefaultScreen(GRAPHICS_DISPLAY);
  root  =XDefaultRootWindow(GRAPHICS_DISPLAY);
  VisualInfo=XDefaultVisual(GRAPHICS_DISPLAY,screen);
  if (LMAXCOLORS>216) 
    LMAXCOLORS=216;
  ColorTable=XDefaultColormap(GRAPHICS_DISPLAY,screen);
  for (i=0;i<40;i++) {
    q.pixel=(unsigned long)i;
    XQueryColor(GRAPHICS_DISPLAY,ColorTable,&q);
    ColorEntry[i]=q;
  }
  ColorTable=XCreateColormap(GRAPHICS_DISPLAY,root,VisualInfo,AllocAll);
  for (i=40;i<255;i++) {
    ColorEntry[i].pixel=(unsigned long)i;
    ColorEntry[i].flags=DoRed | DoGreen | DoBlue;
    ColorEntry[i].red  =0;
    ColorEntry[i].green=0;
    ColorEntry[i].blue =0;
    XAllocColor(GRAPHICS_DISPLAY,ColorTable,&ColorEntry[i]);
  }
  create_standard_colors();
  XStoreColors(GRAPHICS_DISPLAY,ColorTable,ColorEntry,256);
  GCValues.function  =GXcopy;
  GCValues.plane_mask=AllPlanes;
  GCValues.foreground=1;
  GCValues.background=0;
  GRAPHICS_CONTEXT=XCreateGC(GRAPHICS_DISPLAY,root,
                             GCFunction | GCPlaneMask | GCForeground | GCBackground,&GCValues); 
  XInstallColormap(GRAPHICS_DISPLAY,ColorTable);
  XFlush(GRAPHICS_DISPLAY);
}

void closegraph(Display *display=NULL) {
  XUninstallColormap(GRAPHICS_DISPLAY,ColorTable);
  XFreeColormap(GRAPHICS_DISPLAY,ColorTable);
  XFreeGC(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT);
  if (display==NULL) XCloseDisplay(GRAPHICS_DISPLAY);
}

//-------------------------------------------------------------------

int getmaxcolor() { return(LMAXCOLORS); }

int getmaxx(Window window=-MAXINT) { 
  if (window==-MAXINT) return FIND_GW(default_window)->width;
  else return FIND_GW(window)->width;
}

int getmaxy(Window window=-MAXINT) { 
  if (window==-MAXINT) return FIND_GW(default_window)->height;
  else return FIND_GW(window)->height;
}

void getsize(Window window,int &w,int &h) {
  GW *p;

  p=FIND_GW(window);
  w=p->width;
  h=p->height;
}

void getsize(int &w,int &h) { getsize(default_window,w,h); }

void getposition(Window window,int &x,int &y) {
  XWindowAttributes Attr;
  GW *p;

  p=FIND_GW(window);
  XGetWindowAttributes(GRAPHICS_DISPLAY,window,&Attr);
  p->x=Attr.x;
  p->y=Attr.y;
  x=p->x;
  y=p->y;
}

void getposition(int &x,int &y) { getposition(default_window,x,y); }

//-------------------------------------------------------------------

void setcolor(Window window,int color) {
  XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
  FIND_GW(window)->foreground=color;
}

void setcolor(int color) { setcolor(default_window,color); }

int getfgcolor(Window window=-MAXINT) { 
  if (window==-MAXINT) return FIND_GW(default_window)->foreground;
  else return FIND_GW(window)->foreground;
}

void setbkcolor(Window window,int color) {
  XSetWindowBackground(GRAPHICS_DISPLAY,window,color+ColorOfs);
  FIND_GW(window)->background=color;
}

void setbkcolor(int color) { setbkcolor(default_window,color); }

int getbkcolor(Window window=-MAXINT) { 
  if (window==-MAXINT) return FIND_GW(default_window)->background;
  else return FIND_GW(window)->background;
}

void keepratio(Window window,Boolean keepit) {
  GW *p=FIND_GW(window);

  if (keepit) p->ratio=(float)(p->width)/(float)(p->height);
  else p->ratio=-1;
}

void keepratio(Boolean keepit) { keepratio(default_window,keepit); }

//--- moveto --------------------------------------------------------

void moveto(Window window,int x,int y) { SET_pen(window,x,y); }
void moveto(int x,int y) { SET_pen(default_window,x,y); }

//--- putpixel ------------------------------------------------------

void putpixel(Window window,int x,int y,int color=-MAXINT) {
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawPoint(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x,y);
  XFlush(GRAPHICS_DISPLAY);
}

void putpixel(int x,int y,int col=-MAXINT) { putpixel(default_window,x,y,col); }

//-------------------------------------------------------------------

void line(Window window,int x1,int y1,int x2,int y2,int color=-MAXINT) {
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawLine(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x1,y1,x2,y2);	
  XFlush(GRAPHICS_DISPLAY);
}

void line(int x1,int y1,int x2,int y2,int color=-MAXINT) { line(default_window,x1,y1,x2,y2,color); }

void lineto(Window window,int x,int y,int color=-MAXINT) {
  GW *p;
  int px,py;

  p=FIND_GW(window);
  px=p->penx;
  py=p->peny;
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawLine(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,px,py,x,y);	
  XFlush(GRAPHICS_DISPLAY);
  p->penx=x;
  p->peny=y;
}

void lineto(int x,int y,int color=-MAXINT) { lineto(default_window,x,y,color); }

void drawpoly(Window window,int num,polypoint *points,int color=-MAXINT) {
  XPoint *Points;

  Points=(XPoint*)points;
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawLines(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,Points,num,CoordModeOrigin);
  XFlush(GRAPHICS_DISPLAY);
}

void drawpoly(int num,polypoint *points,int color=-MAXINT) { drawpoly(default_window,num,points,color); }

void fillpoly(Window window,int num,polypoint *points,int color=-MAXINT) {

  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XFillPolygon(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,points,num,0,CoordModeOrigin);
  XFlush(GRAPHICS_DISPLAY);                                     // no one knows why == 0
}

void fillpoly(int num,polypoint *points,int color=-MAXINT) { fillpoly(default_window,num,points,color); }

void rectangle(Window window,int xa,int ya,int xb,int yb,int color= -MAXINT) {
  line(window,xa,ya,xb,ya,color);
  line(window,xb,ya,xb,yb,color);
  line(window,xb,yb,xa,yb,color);
  line(window,xa,yb,xa,ya,color);
}

void rectangle(int xa,int ya,int xb,int yb,int color= -MAXINT) { rectangle(default_window,xa,ya,xb,yb,color);}

void bar(Window window,int xa,int ya,int xb,int yb,int color=-MAXINT) {
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XFillRectangle(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,xa,ya,xb-xa,yb-ya);
}

void bar(int xa,int ya,int xb,int yb,int color=-MAXINT) { bar(default_window,xa,ya,xb,yb,color); }

void putrgbpixel(Window window,int x,int y,int r,int g,int b) {
  XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,ConvRGBtoPIX(r,g,b));
  XDrawPoint(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x,y);
  XFlush(GRAPHICS_DISPLAY);
}

void putrgbpixel(int x,int y,int r,int g,int b) { putrgbpixel(default_window,x,y,r,g,b); }

//-------------------------------------------------------------------

unsigned long imagesize(int x1,int y1,int x2,int y2) {
  unsigned long s;

  s=(unsigned long)(x2-x1+1)*(unsigned long)(y2-y1+1)+4;
  return s;
}

void getimage(Window window,int x1,int y1,int x2,int y2,void *bitmap) {
  unsigned short w,h,*d;
  char *data;
  XImage *p;  
  unsigned long i,j,o;

  data=(char*)bitmap;
  d=(unsigned short*)bitmap;
  data++;
  data++;
  data++;
  data++;

  w=x2-x1+1;
  h=y2-y1+1;
  (*d)=w;
  d++;
  (*d)=h;

  p=XCreateImage(GRAPHICS_DISPLAY,VisualInfo,8,ZPixmap,0,data,w,h,8,w);
  (*p)=*XGetImage(GRAPHICS_DISPLAY,window,x1,y1,w,h,0xffffffff,ZPixmap);
  o=p->bytes_per_line-w;
  for (i=0;i<h;i++) for (j=0;j<w;j++) data[i*w+j]=p->data[i*(w+o)+j];
}

void getimage(int x1,int y1,int x2,int y2,void *bitmap) { getimage(default_window,x1,y1,x2,y2,bitmap); }

void putimage(Window window,int x,int y,void *bitmap,int /*flag*/ ) {
  XImage *p;
  unsigned short w,h,*d;
  char *data;

  data=(char*)bitmap;
  d=(unsigned short*)bitmap;
  data++;
  data++;
  data++;
  data++;
  w=(*d);
  d++;
  h=(*d);
  p=XCreateImage(GRAPHICS_DISPLAY,VisualInfo,8,ZPixmap,0,data,w,h,8,w);
  XPutImage(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,p,0,0,x,y,w,h);
  XFlush(GRAPHICS_DISPLAY);
}

void putimage(int x,int y,void *bitmap,int flag) { putimage(default_window,x,y,bitmap,flag); }
/*
void putimage(Window window,int x,int y,Impvaru *image) {
  XImage *p;
  int w,h;

  w=(int)image->xsize;
  h=(int)image->ysize;
  p=XCreateImage(GRAPHICS_DISPLAY,VisualInfo,8,ZPixmap,0,image->datptr,w,h,8,w);
  XPutImage(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,p,0,0,x,y,w,h);
  XFlush(GRAPHICS_DISPLAY);
}
*/
//-----------------------------------------------------------------------

void outtextxy(Window window,int x,int y,char *s,int color=-MAXINT) {
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawString(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x,y,s,(int)strlen(s));
}

void outtextxy(int x,int y,char *s,int color=-MAXINT) { outtextxy(default_window,x,y,s,color); }

void outtext(Window window,char *s,int color=-MAXINT) {
  GW* p;

  p=FIND_GW(window);
  if (p!=NULL) outtextxy(window,p->penx,p->peny,s,color);
}


void outtext(char *s,int color=-MAXINT) { outtext(default_window,s,color); }

//-----------------------------------------------------------------------

void arc(Window window,int x,int y,int startangle,int stopangle,int r,int color=-MAXINT) {
  if (color>-MAXINT) XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,color+ColorOfs);
    else XSetForeground(GRAPHICS_DISPLAY,GRAPHICS_CONTEXT,FIND_GW(window)->foreground+ColorOfs);
  XDrawArc(GRAPHICS_DISPLAY,window,GRAPHICS_CONTEXT,x-r,y-r,2*r,2*r,startangle*64,stopangle*64);
}

void arc(int x,int y,int startangle,int stopangle,int r,int color=-MAXINT) { 
  arc(default_window,x,y,startangle,stopangle,r,color);
}

//-----------------------------------------------------------------------

void circle(Window window,int x,int y,int r,int color=-MAXINT) { arc(window,x,y,r,0,360,color); }

void circle(int x,int y,int r,int color=-MAXINT) { circle(default_window,x,y,r,color); }
















