/*--------------------------------------------------------------------
  File:        xgraphics.h
  Author:      Nils Schmeisser
  ----------------------------------------------------------------------
  Description: Graphics module for X applications
  ----------------------------------------------------------------------
  Notes:       
  Date:        3.12.1993
  ----------------------------------------------------------------------

   Copyright (C) 1993 Nils Schmeisser

This file is part of libxapp.

libxapp is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

libxapp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with libxapp; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#define XTFUNCPROTO

#include <impdef.h>
#include <impdat.h>

#include <values.h>
#include <Xm/Xm.h>

typedef void (*XGraphicsRedrawFunction)(Window,int);
typedef XPoint polypoint;

extern int xBLACK;
extern int xBLUE;
extern int xGREEN;
extern int xCYAN;
extern int xRED;
extern int xMAGENTA;
extern int xBROWN;
extern int xLIGHTGRAY;
extern int xDARKGRAY;
extern int xLIGHTBLUE;
extern int xLIGHTGREEN;
extern int xLIGHTCYAN;
extern int xLIGHTRED;
extern int xLIGHTMAGENTA;
extern int xYELLOW;
extern int xWHITE;

extern void initgraph(Display* =NULL);
extern void closegraph(Display* =NULL);
extern Display *getgraphicsdisplay();
extern void setgraphicsdisplay(Display *dpy);

extern Window openwindow(char *name,int x,int y,int sizex,int sizey,Widget parent,Widget *da=NULL);
extern Window openwindow(char *name,int x,int y,int sizex,int sizey,Widget parent,
                         XGraphicsRedrawFunction f,Widget *da=NULL);
extern void closewindow(Window window);
extern void setwindow(Window window);
extern void setredrawfunction(Window window,XGraphicsRedrawFunction f);

//--- full calling style --------------------------------------------------------------------

extern void cleardevice(Window =-MAXINT);

extern int getmaxcolor();
extern void setcolor(Window window,int color);
extern int getfgcolor(Window window=-MAXINT);
extern void setbkcolor(Window window,int color);
extern int getbkcolor(Window window=-MAXINT);
extern void setrgbpalette(int colornum, int red, int green, int blue);
void keepratio(Window window,Boolean keepit);

extern int getmaxx(Window =-MAXINT);
extern int getmaxy(Window =-MAXINT);
extern void getsize(Window window,int &w,int &h);
extern void getposition(Window window,int &x,int &y);

extern void moveto(Window window,int x,int y);

extern void putpixel(Window window,int x,int y,int =-MAXINT);
extern void putrgbpixel(Window window,int x,int y,int r,int g,int b);

extern void line(Window window,int x1,int y1,int x2,int y2,int =-MAXINT);
extern void lineto(Window window,int x,int y,int =-MAXINT);
extern void drawpoly(Window window,int num,polypoint *points,int =-MAXINT);
extern void fillpoly(Window window,int num,polypoint *points,int =-MAXINT);

extern void rectangle(Window window,int xa,int ya,int xb,int yb,int = -MAXINT);
extern void bar(Window window,int xa,int ya,int xb,int yb,int = -MAXINT);

extern void arc(Window window,int x,int y,int startangle,int stopangle,int r,int =-MAXINT);
extern void circle(Window window,int x,int y,int r,int =-MAXINT);

extern unsigned long imagesize(int x1,int y1,int x2,int y2);
extern void getimage(Window window,int x1,int y1,int x2,int y2,void *bitmap);
extern void putimage(Window window,int x,int y,void *bitmap,int flag);

extern void outtextxy(Window window,int x,int y,char *s,int =-MAXINT);
extern void outtext(Window window,char *s,int =-MAXINT);

//--- without window calling style --------------------------------------------------

extern void setcolor(int color);
extern void setbkcolor(int color);
void keepratio(Boolean keepit);

extern void getsize(int &w,int &h);
extern void getposition(int &x,int &y);

extern void moveto(int x,int y);

extern void putpixel(int x,int y,int =-MAXINT);
extern void putrgbpixel(int x,int y,int r,int g,int b);

extern void line(int x1,int y1,int x2,int y2,int =-MAXINT);
extern void lineto(int x,int y,int =-MAXINT);
extern void drawpoly(int num,polypoint *points,int =-MAXINT);
extern void fillpoly(int num,polypoint *points,int =-MAXINT);

extern void rectangle(int xa,int ya,int xb,int yb,int = -MAXINT);
extern void bar(int xa,int ya,int xb,int yb,int = -MAXINT);

extern void arc(int x,int y,int startangle,int stopangle,int r,int =-MAXINT);
extern void circle(int x,int y,int r,int =-MAXINT);

extern void getimage(int x1,int y1,int x2,int y2,void *bitmap);
extern void putimage(int x,int y,void *bitmap,int flag);
extern void outtextxy(int x,int y,char *s,int =-MAXINT);
extern void outtext(char *s,int =-MAXINT);


extern void putimage(Window window,int x,int y,Impvaru *image);





